#include "ast3d.h"

const int JOYSTICK_PORT = 0x0201;
const int STICK_1_X_BIT = 0x0001;
const int STICK_1_Y_BIT = 0x0002;
const int STICK_2_X_BIT = 0x0004;
const int STICK_2_Y_BIT = 0x0008;

/*
** AstGameport::AstGameport( int style )
**
** This is the AstGameport class' constructor.  The parameter is
** a bitfield defining which stick axes should be checked when
** the gameport "Update()" method is called.  It also initializes
** the stick variables.
*/
AstGameport::AstGameport( int s )
{
   style = s;
   portmask = 0;

   if ( s & gpStick1 ) {
      portmask |= STICK_1_X_BIT;
      portmask |= STICK_1_Y_BIT;
   }
   if ( s & gpRCS ) {
      portmask |= STICK_2_X_BIT;
   }
   if ( s & gpWCS ) {
      portmask |= STICK_2_Y_BIT;
   }

   /*
   ** Zero out stick variables
   */
   for ( int i = 0; i < 2; i++ ) {
      memset( &( stick[i] ), 0, sizeof( AstGamePortState ) );
   }
}

/*
** void AstGameport::AdjustForCalibration( void )
**
** This routine adjusts the raw values read by the gameport against
** calibration values.  This routine is automatically called up
** the public Update() function.
*/
void AstGameport::AdjustForCalibration( void )
{
   if ( style & gpStick1 ) {

      stick[0].rawx -= stick[0].lx;
      stick[0].rawy -= stick[0].ly;

      if ( ( stick[0].rawx >= stick[0].dead_zone_lx ) &&
           ( stick[0].rawx <= stick[0].dead_zone_rx ) ) {
         stick[0].x = 0;
      }
      else {
         if ( stick[0].rawx < stick[0].dead_zone_lx ) {
            stick[0].x = AstFixMulFix( AstConvertToFix( stick[0].rawx ),
                                       stick[0].scalex_l );
            stick[0].x -= AstConvertToFix( 1 );
            if ( stick[0].x < AstConvertToFix( -1 ) )
               stick[0].x = AstConvertToFix( -1 );
         }
         else {
            stick[0].rawx -= stick[0].cx;
            stick[0].x = AstFixMulFix( AstConvertToFix( stick[0].rawx ),
                                       stick[0].scalex_r );
            if ( stick[0].x > AstConvertToFix( 1 ) )
               stick[0].x = AstConvertToFix( 1 );
         }
      }

      if ( ( stick[0].rawy >= stick[0].dead_zone_ly ) &&
           ( stick[0].rawy <= stick[0].dead_zone_ry ) ) {
         stick[0].y = 0;
      }
      else {
         if ( stick[0].rawy < stick[0].dead_zone_ly ) {
            stick[0].y = AstFixMulFix( AstConvertToFix( stick[0].rawy ),
                                       stick[0].scaley_t );
            stick[0].y -= AstConvertToFix( 1 );
            if ( stick[0].y < AstConvertToFix( -1 ) )
               stick[0].y = AstConvertToFix( -1 );
         }
         else {
            stick[0].rawy -= stick[0].cy;
            stick[0].y = AstFixMulFix( AstConvertToFix( stick[0].rawy ),
                                       stick[0].scaley_b );
            if ( stick[0].y > AstConvertToFix( 1 ) )
               stick[0].y = AstConvertToFix( 1 );
         }
      }
   }

   if ( style & gpRCS ) {
      stick[1].rawx -= stick[1].lx;

      if ( ( stick[1].rawx >= stick[1].dead_zone_lx ) &&
           ( stick[1].rawx <= stick[1].dead_zone_rx ) ) {
         stick[1].x = 0;
      }
      else {
         if ( stick[1].rawx < stick[1].dead_zone_lx ) {
            stick[1].x = AstFixMulFix( AstConvertToFix( stick[1].rawx ),
                                       stick[1].scalex_l );
            stick[1].x -= AstConvertToFix( 1 );
            if ( stick[1].x < AstConvertToFix( -1 ) )
               stick[1].x = AstConvertToFix( -1 );
         }
         else {
            stick[1].rawx -= stick[1].cx;
            stick[1].x = AstFixMulFix( AstConvertToFix( stick[1].rawx ),
                                       stick[1].scalex_r );
            if ( stick[1].x > AstConvertToFix( 1 ) )
               stick[1].x = AstConvertToFix( 1 );
         }
      }
   }

   if ( style & gpWCS ) {
      stick[1].rawy -= stick[1].ly;

      stick[1].y = AstConvertToFix( 1 ) -
                   AstFixMulFix( AstConvertToFix( stick[1].rawy ),
                                 stick[1].scaley_t );
      if ( stick[1].y < 0 )
         stick[1].y = 0;
      else if ( stick[1].y > AstConvertToFix( 1 ) )
         stick[1].y = AstConvertToFix( 1 );
   }
}

/*
** void AstGameport::Calibrate( void )
**
** Calibrates the gameport.  Calibration prompting assumes text mode
** so we cannot be in a graphics mode when this code is executed.
*/
void AstGameport::Calibrate( void )
{
  if ( style & gpStick1 ) {
     CalibrateJoystick();
  }
  if ( style & gpRCS ) {
     CalibrateRCS();
  }
  if ( style & gpWCS ) {
     CalibrateWCS();
  }
}

void AstGameport::CalibrateJoystick( void )
{

   puts( "Center the joystick and press a button" );
   WaitForButtonPress();
   stick[0].cx = stick[0].rawx;
   stick[0].cy = stick[0].rawy;
   WaitForButtonRelease();

   puts( "Move joystick to the upper left corner and press a button" );
   WaitForButtonPress();
   stick[0].lx = stick[0].rawx;
   stick[0].ly = stick[0].rawy;
   WaitForButtonRelease();

   puts( "Move joystick to the lower right corner and press a button" );
   WaitForButtonPress();
   stick[0].rx = stick[0].rawx;
   stick[0].ry = stick[0].rawy;
   WaitForButtonRelease();

   /*
   ** Transform all of our points so that 0,0 is the top left corner,
   ** stick[0].c? is the center point, and stick[0].r? is the bottom right.
   */
   stick[0].cx -= stick[0].lx;
   stick[0].cy -= stick[0].ly;
   stick[0].rx -= stick[0].lx;
   stick[0].ry -= stick[0].ly;

   /*
   ** Compute dead zone area
   */
   const float deadzone = .1;

   stick[0].dead_zone_lx = stick[0].cx - stick[0].cx * deadzone;
   stick[0].dead_zone_rx = stick[0].cx + stick[0].cx * deadzone;
   stick[0].dead_zone_ly = stick[0].cy - stick[0].cy * deadzone;
   stick[0].dead_zone_ry = stick[0].cy + stick[0].cy * deadzone;

   /*
   ** Compute scaling factors
   */
   float range;

   // Determine range of left hand side
   range = 1.0 / ( stick[0].cx );
   stick[0].scalex_l = AstConvertToFix( range );

   // Determine range of right hand side
   range = 1.0 / ( stick[0].rx - stick[0].cx );
   stick[0].scalex_r = AstConvertToFix( range );

   // Determine range of top side
   range = 1.0 / stick[0].cy;
   stick[0].scaley_t = AstConvertToFix( range );

   // Determine range of bottom side
   range = 1.0 / ( stick[0].ry - stick[0].cy );
   stick[0].scaley_b = AstConvertToFix( range );

}

void AstGameport::CalibrateRCS( void )
{
   puts( "Center the rudder pedals and press a key" );
   getch();
   ReadPort();
   stick[1].cx = stick[1].rawx;

   puts( "Completely depress the left pedal and press a key" );
   getch();
   ReadPort();
   stick[1].lx = stick[1].rawx;

   puts( "Completely depress the right pedal and press a key" );
   getch();
   ReadPort();
   stick[1].rx = stick[1].rawx;

   /*
   ** Recenter CX, CY, RX, and RY
   */
   stick[1].cx -= stick[1].lx;
   stick[1].rx -= stick[1].lx;

   /*
   ** Compute dead zone area
   */
   const float deadzone = .2;

   stick[1].dead_zone_lx = stick[1].cx - stick[1].cx * deadzone;
   stick[1].dead_zone_rx = stick[1].cx + stick[1].cx * deadzone;

   /*
   ** Compute scaling factors
   */
   float range;

   range = 1.0 / ( stick[1].cx );
   stick[1].scalex_l = AstConvertToFix( range );

   range = 1.0 / ( stick[1].rx - stick[1].cx );
   stick[1].scalex_r = AstConvertToFix( range );
}

void AstGameport::CalibrateWCS( void )
{
   puts( "WARNING:  Please make sure that your WCS is set to ANALOG mode\n" );
   puts( "Press your throttle all the way forward and press a key" );
   getch();
   ReadPort();
   stick[1].ly = stick[1].rawy;

   puts( "Pull your throttle completely back and press a key" );
   getch();
   ReadPort();
   stick[1].ry = stick[1].rawy;

   /*
   ** Recenter Cy, CY, Ry, and RY
   */
   stick[1].ry -= stick[1].ly;

   /*
   ** Compute dead zone area
   */
   const float deadzone = .2;

   stick[1].dead_zone_ly = stick[1].cy - stick[1].cy * deadzone;
   stick[1].dead_zone_ry = stick[1].cy + stick[1].cy * deadzone;

   /*
   ** Compute scaling factors
   */
   float range;

   // Determine range
   range = 1.0 / ( stick[1].ry );
   stick[1].scaley_t = AstConvertToFix( range );
}

/*
** void AstGameport::ReadPort( void )
**
** This routine reads the joystick port, located at port address
** 0x201.  The joystick port is kind of weird, in that the way you
** read values from it is by triggering some one-shot counters and
** timing the amount of time it takes for the counters to count
** out.  Primitive, and unreliable, but that's the way it is.
**
** Values read for the various sticks consist of arbitrary (and CPU-
** dependent) values which must be adjusted for calibration.
*/
void AstGameport::ReadPort( void )
{
int x1, y1, x2, y2, c;

   /*
   ** Disable interrupts since an interrupt can skew our timing
   */
   DISABLE();

   /*
   ** Trigger game port one shot by writing out an arbitrary value
   ** to the gameport.
   */
   OUTPORTB( JOYSTICK_PORT, 0xff );

   /*
   ** Do the initial read
   */
   c = buttons = INPORTB( JOYSTICK_PORT );

   /*
   ** Keep reading until the one shot counts out
   */
   for ( int k = 0; ( c & portmask ) != 0; k++ ) {
       if ( ( c & STICK_1_X_BIT ) != 0 )
          x1 = k;
       if ( ( c & STICK_1_Y_BIT ) != 0)
          y1 = k;
       if ( ( c & STICK_2_X_BIT ) != 0 )
          x2 = k;
       if ( ( c & STICK_2_Y_BIT ) != 0 )
          y2 = k;
       c = INPORTB( JOYSTICK_PORT );
   }
   /*
   ** Reenable interrupts
   */
   ENABLE();

   /*
   ** Fill in the values
   */
   stick[0].rawx = x1;
   stick[0].rawy = y1;
   stick[1].rawx = x2;
   stick[1].rawy = y2;

   /*
   ** Set our button value
   */
   buttons = ( ( ~buttons ) >> 4 ) & 0x000f;
}

void AstGameport::WaitForButtonPress( void )
{
   ReadPort();
   while ( !buttons )
      ReadPort();
}

void AstGameport::WaitForButtonRelease( void )
{
   ReadPort();
   while ( buttons )
      ReadPort();
}
