#include "ast3d.h"

AstVector operator - ( const AstVector &v )
{
   return AstVector( -v.X, -v.Y, -v.Z );
}

/*
** AstVector AstVector::Cross( const AstVector &b ) const
**
** Computes the cross product of this vector and the passed
** vector.  The cross product of two 3D vectors will result
** in a vector normal (perpendicular) to the plane defined
** by the two initial 3D vectors.
*/
AstVector AstVector::Cross( const AstVector &b ) const
{
AstVector result;

   result.X = AstFixMulFix( Y, b.Z ) - AstFixMulFix( Z, b.Y );
   result.Y = AstFixMulFix( Z, b.X ) - AstFixMulFix( X, b.Z );
   result.Z = AstFixMulFix( X, b.Y ) - AstFixMulFix( Y, b.X );

   return result;
}

/*
** AstFixedpoint_t AstVector::Dot ( const AstVector & ) const
**
** This routine computes the dot product between two fixed point
** vectors.  The dot product of two vectors determines the angle
** between them and is computed as:
**
**     A.X * B.X + A.Y * B.Y + A.Z * B.Z
**
** The result is the cosine of the angle between the two vectors,
** and is useful for shading (where the angle determines how
** directly the light hits a surface) and backface culling
** (where the angle determines if the viewer and a face are facing
** each other).
*/
AstFixedpoint_t AstVector::Dot( const AstVector &b ) const
{
   return ( AstFixMulFix( X, b.X ) +
            AstFixMulFix( Y, b.Y ) +
            AstFixMulFix( Z, b.Z ) );
}

/*
** void AstVector::operator *= ( AstFixedpoint_t scaler )
**
** This routine overloads the "*=" operator so that a vector
** can be multiplied by a fixed point scaler easily.  This
** routine will scale the vector.
*/
void AstVector::operator *= ( AstFixedpoint_t s )
{
   X = AstFixMulFix( X, s );
   Y = AstFixMulFix( Y, s );
   Z = AstFixMulFix( Z, s );
}

/*
** AstVector operator * ( const AstVector &v, AstMatrix4x3 &m )
**
** This routine overloads the "*" operator so that a
** vector can be multiplied by a 4x3 matrix easily.  This routine
** will transform the vector and return the new vector.
*/
AstVector operator * ( const AstVector &v, const AstMatrix4x3 &m )
{
AstVector result;

  result.X = AstFixMulFix( m[0][0], v.X ) +
             AstFixMulFix( m[1][0], v.Y ) +
             AstFixMulFix( m[2][0], v.Z ) +
             m[3][0];

  result.Y = AstFixMulFix( m[0][1], v.X ) +
             AstFixMulFix( m[1][1], v.Y ) +
             AstFixMulFix( m[2][1], v.Z ) +
             m[3][1];

  result.Z = AstFixMulFix( m[0][2], v.X ) +
             AstFixMulFix( m[1][2], v.Y ) +
             AstFixMulFix( m[2][2], v.Z ) +
             m[3][2];

  return result;
}

/*
** AstVector operator * ( const AstVector &v, AstMatrix3x3 &m )
** void AstVector::operator *= ( AstMatrix3x3 &m )
**
** This routine overloads the "*" and "*=" operator so that a
** vector can be multiplied by a 3x3 matrix easily.  This routine
** will transform (since this is a 3x3 matrix, the transformation
** will in all likelihood be a rotation) the passed vector and
** return the transformed vector.
*/
AstVector operator * ( const AstVector &v, const AstMatrix3x3 &m )
{
AstVector result;

  result.X = AstFixMulFix( m[0][0], v.X ) +
             AstFixMulFix( m[1][0], v.Y ) +
             AstFixMulFix( m[2][0], v.Z );

  result.Y = AstFixMulFix( m[0][1], v.X ) +
             AstFixMulFix( m[1][1], v.Y ) +
             AstFixMulFix( m[2][1], v.Z );

  result.Z = AstFixMulFix( m[0][2], v.X ) +
             AstFixMulFix( m[1][2], v.Y ) +
             AstFixMulFix( m[2][2], v.Z );

  return result;
}

void AstVector::operator *= ( const AstMatrix3x3 &m )
{
AstFixedpoint_t save_x, save_y, save_z;

  save_x = X;
  save_y = Y;
  save_z = Z;

  X = AstFixMulFix( m[0][0], save_x ) +
      AstFixMulFix( m[1][0], save_y ) +
      AstFixMulFix( m[2][0], save_z );

  Y = AstFixMulFix( m[0][1], save_x ) +
      AstFixMulFix( m[1][1], save_y ) +
      AstFixMulFix( m[2][1], save_z );

  Z = AstFixMulFix( m[0][2], save_x ) +
      AstFixMulFix( m[1][2], save_y ) +
      AstFixMulFix( m[2][2], save_z );
}

/*
** AstFixedpoint_t AstVector::Magnitude( void ) cpmst
**
** This function calculates the length, or magnitude, of a
** vector.  This function is required when normalizing the
** vector.  The magnitude of a vector is equal to the distance
** from (0,0,0) to the vector, or:
**
**     sqrt( X*X + Y*Y + Z*Z )
*/
AstFixedpoint_t AstVector::Magnitude( void ) const
{
float f;

   f  = AstConvertToFloat( X ) * AstConvertToFloat( X ) +
        AstConvertToFloat( Y ) * AstConvertToFloat( Y ) +
        AstConvertToFloat( Z ) * AstConvertToFloat( Z );

   return ( AstConvertToFix( sqrt( f ) ) );
}

int AstVector::MagnitudeSquared( void ) const
{
  return AstConvertToInt(   AstFixMulFix( X, X )
                          + AstFixMulFix( Y, Y )
                          + AstFixMulFix( Z, Z ) );
}

/*
** AstFixedpoint_t AstVector::Normalize( void )
**
** This function normalizes a vector, which means that its
** length is reduced to 1.  Such a vector is known as a
** normalized or unit vector.
*/
void AstVector::Normalize( void )
{
AstFixedpoint_t magnitude = Magnitude();

   X = AstFixDivFix( X, magnitude );
   Y = AstFixDivFix( Y, magnitude );
   Z = AstFixDivFix( Z, magnitude );
}

